import logging
from argparse import ArgumentParser


def execute():
    parser = ArgumentParser(
        prog='dbgear',
        description='database management tools'
    )
    parser.add_argument(
        '--project',
        default='database',
        help='please specify the folder for the project.')

    sub = parser.add_subparsers(dest='command', help='sub-command help')

    apply_parser = sub.add_parser('apply', help='apply help')
    apply_parser.add_argument(
        'deploy',
        help='target deployment.')
    apply_parser.add_argument(
        'env',
        help='target environment.')
    apply_parser.add_argument(
        '--database',
        help='target database.')
    apply_parser.add_argument(
        '--target',
        help='target table.')
    apply_parser.add_argument(
        '--all',
        choices=['drop', 'delta'],
        help='apply all tables. Specify "drop" to drop database before applying, or "delta" to apply only the changes since the last deployment.'
    )
    apply_parser.add_argument(
        '--no-restore',
        action='store_true',
        help='skip all data operations (initial data loading and backup restoration)'
    )
    apply_parser.add_argument(
        '--restore-only',
        action='store_true',
        help='restore initial data only without recreating schema'
    )
    apply_parser.add_argument(
        '--patch',
        help='patch file to execute instead of restore'
    )
    apply_parser.add_argument(
        '--backup-key',
        help='backup table timestamp key (YYYYMMDDHHMMSS). Only valid with --restore-only'
    )

    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    logging.info(args)

    if args.command is None:
        parser.print_help()
        return

    # For other commands, load project and required modules
    from .models.project import Project
    from . import operations

    project = Project.load(args.project)

    if args.command == 'apply':
        if not args.all and args.target is None:
            logging.error('please specify --target or --all')
            return

        # Validate backup-key option
        if args.backup_key and not args.restore_only:
            logging.error('--backup-key can only be used with --restore-only')
            return

        # Validate backup-key format
        if args.backup_key:
            import re
            if not re.match(r'^\d{14}$', args.backup_key):
                logging.error('--backup-key must be in YYYYMMDDHHMMSS format')
                return

        operations.apply(
            project,
            args.env,
            args.database,
            args.target,
            args.all,
            args.deploy,
            args.no_restore,
            args.restore_only,
            args.patch,
            args.backup_key
        )



"""
Patch processing module for DBGear.

Handles YAML patch file parsing and SQL generation for custom data restoration.
"""

import yaml
import logging
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


class PatchConfig:
    """Represents a patch configuration loaded from YAML."""

    def __init__(self, name: str, select: List[str], where: Optional[str] = None):
        self.name = name
        self.select = select
        self.where = where

    @classmethod
    def from_dict(cls, data: Dict) -> 'PatchConfig':
        """Create PatchConfig from dictionary."""
        if 'name' not in data:
            raise ValueError("Patch config must have 'name' field")
        if 'select' not in data:
            raise ValueError("Patch config must have 'select' field")

        return cls(
            name=data['name'],
            select=data['select'],
            where=data.get('where')
        )

    @classmethod
    def load_from_file(cls, file_path: str) -> 'PatchConfig':
        """Load patch configuration from YAML file."""
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                data = yaml.safe_load(f)

            return cls.from_dict(data)
        except FileNotFoundError:
            raise ValueError(f"Patch file not found: {file_path}")
        except yaml.YAMLError as e:
            raise ValueError(f"Invalid YAML in patch file: {e}")


def generate_patch_sql(env: str, patch_config: PatchConfig, backup_key: str) -> str:
    """
    Generate INSERT SQL from patch configuration.

    Args:
        env: Database environment name
        patch_config: Patch configuration
        backup_key: Backup table timestamp key (YYYYMMDDHHMMSS)

    Returns:
        Generated INSERT SQL string
    """
    table_name = patch_config.name
    backup_table = f"bak_{table_name}_{backup_key}"

    # Build SELECT columns
    select_columns = ",\n  ".join(patch_config.select)

    # Build WHERE clause if specified
    where_clause = ""
    if patch_config.where:
        where_clause = f"\nWHERE {patch_config.where}"

    # Generate final SQL
    sql = f"""INSERT INTO {env}.{table_name}
SELECT
  {select_columns}
FROM {env}.{backup_table}{where_clause}"""

    return sql


def validate_patch_config(patch_config: PatchConfig) -> List[str]:
    """
    Validate patch configuration and return list of warnings/errors.

    Returns:
        List of validation messages (empty if valid)
    """
    errors = []

    if not patch_config.name:
        errors.append("Table name cannot be empty")

    if not patch_config.select:
        errors.append("Select columns cannot be empty")

    # Check for common SQL injection patterns in WHERE clause
    if patch_config.where:
        dangerous_patterns = [';', '--', '/*', '*/', 'DROP', 'DELETE', 'UPDATE']
        where_upper = patch_config.where.upper()
        for pattern in dangerous_patterns:
            if pattern in where_upper:
                errors.append(f"Potentially dangerous pattern '{pattern}' found in WHERE clause")
                break

    return errors



"""Database stored procedure operations."""

from . import engine
from .templates.mysql import template_engine
from ..models.procedure import Procedure


def is_exist(conn, env: str, procedure: Procedure):
    """Check if stored procedure exists"""
    sql = template_engine.render('mysql_check_procedure_exists')
    result = engine.select_one(conn, sql, {
        'env': env,
        'procedure_name': procedure.procedure_name,
        'routine_type': 'FUNCTION' if procedure.is_function else 'PROCEDURE'
    })
    return result is not None


def drop(conn, env: str, procedure: Procedure):
    """Drop stored procedure or function"""
    if procedure.is_function:
        sql = template_engine.render('mysql_drop_function', env=env, procedure_name=procedure.procedure_name)
    else:
        sql = template_engine.render('mysql_drop_procedure', env=env, procedure_name=procedure.procedure_name)
    engine.execute(conn, sql)


def create(conn, env: str, procedure: Procedure):
    """Create stored procedure or function"""
    if procedure.is_function:
        sql = template_engine.render('mysql_create_function', env=env, procedure=procedure)
    else:
        sql = template_engine.render('mysql_create_procedure', env=env, procedure=procedure)
    engine.execute(conn, sql)



import pydantic

from .base import BaseSchema
from .notes import Note
from .notes import NoteManager


class ProcedureParameter(BaseSchema):
    """Stored procedure parameter definition"""
    parameter_name: str
    parameter_type: str  # IN | OUT | INOUT
    data_type: str  # VARCHAR(255), INT, etc.
    default_value: str | None = None


class Procedure(BaseSchema):
    """Stored procedure definition"""
    procedure_name: str = pydantic.Field(exclude=True)
    display_name: str
    parameters: list[ProcedureParameter] = pydantic.Field(default_factory=list)
    return_type: str | None = None  # For functions (NULL for procedures)
    body: str  # SQL procedure body
    language: str = "SQL"  # SQL, PLPGSQL for future database support
    deterministic: bool = False
    reads_sql_data: bool = True
    modifies_sql_data: bool = False
    security_type: str = "DEFINER"  # DEFINER | INVOKER
    notes_: list[Note] = pydantic.Field(default_factory=list, alias='notes')

    @property
    def notes(self) -> NoteManager:
        return NoteManager(self.notes_)

    @property
    def is_function(self) -> bool:
        """Returns True if this is a function (has return type), False if procedure"""
        return self.return_type is not None


class ProcedureManager:
    """Manager for stored procedure collections"""

    def __init__(self, procedures: dict[str, Procedure]):
        self.procedures = procedures

    def __getitem__(self, procedure_name: str) -> Procedure:
        return self.procedures[procedure_name]

    def __iter__(self):
        yield from self.procedures.values()

    def __len__(self) -> int:
        return len(self.procedures)

    def __contains__(self, procedure_name: str) -> bool:
        return procedure_name in self.procedures

    def append(self, procedure: Procedure) -> None:
        if procedure.procedure_name in self.procedures:
            raise ValueError(f"Procedure '{procedure.procedure_name}' already exists")
        self.procedures[procedure.procedure_name] = procedure

    def remove(self, procedure_name: str) -> None:
        if procedure_name not in self.procedures:
            raise KeyError(f"Procedure '{procedure_name}' does not exist")
        del self.procedures[procedure_name]



# レイアウト
LAYOUT_TABLE = 'table'
LAYOUT_MATRIX = 'matrix'
LAYOUT_SINGLE = 'single'

# # カラム定義
# BIND_TYPE_FIXED = 'fixed'
# BIND_TYPE_CALL = 'call'
# BIND_TYPE_EXTEND = 'extend'
# # カラム定義（特殊タイプ）
# BIND_TYPE_REFS = 'refs'
# BIND_TYPE_BLANK = ''

# # callのタイプ
# CALL_TYPE_UUID = 'uuid'
# CALL_TYPE_ULID = 'ulid'

# # FIXME セルタイプ
# FIELD_TYPE_STRING = 'string'
# FIELD_TYPE_SELECTABLE = 'singleSelect'
# FIELD_TYPE_SYSTEM = 'system'

# 同期モード
SYNC_MODE_DROP_CREATE = 'drop_create'
SYNC_MODE_MANUAL = 'manual'

# フィールド幅
DEFAULT_WIDTH = 150

SETTING_TYPE_COLUMN = 'column'
SETTING_TYPE_REF = 'ref'

# データソースのデータタイプ
DATATYPE_YAML = 'yaml'
DATATYPE_XLSX = 'xlsx'
DATATYPE_PYTHON = 'python'



from logging import getLogger
from datetime import datetime

from .dbio import engine
from .dbio import database
from .dbio import table
from .dbio import view
from .dbio import trigger
from .dbio import procedure

from .models.project import Project
from .models.mapping import Mapping
from .models.schema import Schema
from .utils import const

logger = getLogger(__name__)


class Operation:

    def __init__(self, project: Project, env: str, database: str, deploy: str, backup_key: str = None):
        self.project = project
        self.environ = project.envs[env]
        self.database = database

        self.conn = engine.get_connection(self.environ.deployments[deploy])
        # backup_keyが指定されている場合はそれを使用、そうでなければ現在時刻
        self.ymd = backup_key if backup_key else datetime.strftime(datetime.now(), '%Y%m%d%H%M%S')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.conn.close()

    def create_database(self, map: Mapping, all: str):
        if all == 'drop':
            logger.info(f'database {map.instance_name}')
            if database.is_exist(self.conn, map.instance_name):
                database.drop(self.conn, map.instance_name)
            database.create(self.conn, map.instance_name)
        else:
            # 差分更新または個別指定の場合で、データベースが存在しない場合は作成する。
            if not database.is_exist(self.conn, map.instance_name):
                logger.info(f'database {map.instance_name} was created.')
                database.create(self.conn, map.instance_name)

    def create_table(self, map: Mapping, schema: Schema, all: str, target: str, restore_only: bool = False):
        if restore_only:
            logger.info('restore-only mode: skipping schema creation')
            return

        for tbl in schema.tables:
            if not all and target != tbl.table_name:
                continue
            # テーブルが存在しない場合は作成する。
            if not table.is_exist(self.conn, map.instance_name, tbl):
                logger.info(f'table {map.instance_name}.{tbl.table_name} was created.')
                table.create(self.conn, map.instance_name, tbl)
                continue
            # データのバックアップ
            logger.info(f'backup {map.instance_name}.{tbl.table_name}')
            table.backup(self.conn, map.instance_name, tbl, self.ymd)
            # テーブルの再作成
            logger.info(f'drop & create table {map.instance_name}.{tbl.table_name}')
            table.drop(self.conn, map.instance_name, tbl)
            table.create(self.conn, map.instance_name, tbl)

        for vw in schema.views:
            if not all and target != vw.view_name:
                continue
            # ビューが存在しない場合は作成する。
            if not view.is_exist(self.conn, map.instance_name, vw):
                logger.info(f'view {map.instance_name}.{vw.view_name} was created.')
                view.create(self.conn, map.instance_name, vw)
            else:
                # ビューの再作成
                logger.info(f'drop & create view {map.instance_name}.{vw.view_name}')
                view.drop(self.conn, map.instance_name, vw)
                view.create(self.conn, map.instance_name, vw)

        for tr in schema.triggers:
            if not all and target != tr.trigger_name:
                continue
            # トリガーが存在しない場合は作成する。
            if not trigger.is_exist(self.conn, map.instance_name, tr):
                logger.info(f'trigger {map.instance_name}.{tr.trigger_name} was created.')
                trigger.create(self.conn, map.instance_name, tr)
            else:
                # トリガーの再作成
                logger.info(f'drop & create trigger {map.instance_name}.{tr.trigger_name}')
                trigger.drop(self.conn, map.instance_name, tr)
                trigger.create(self.conn, map.instance_name, tr)

        for proc in schema.procedures:
            if not all and target != proc.procedure_name:
                continue
            # プロシージャが存在しない場合は作成する。
            if not procedure.is_exist(self.conn, map.instance_name, proc):
                logger.info(f'procedure {map.instance_name}.{proc.procedure_name} was created.')
                procedure.create(self.conn, map.instance_name, proc)
            else:
                # プロシージャの再作成
                logger.info(f'drop & create procedure {map.instance_name}.{proc.procedure_name}')
                procedure.drop(self.conn, map.instance_name, proc)
                procedure.create(self.conn, map.instance_name, proc)

    def insert_data(self, map: Mapping, schema: Schema, all: bool, target: str, no_restore: bool = False, patch_file: str = None):
        if no_restore:
            # no_restore が指定されている場合は、初期データ投入もバックアップ復元もスキップ
            logger.info('no-restore mode: skipping all data operations')
            return

        # データ投入の順序を決定
        if all:
            # 全体指定時は依存関係を考慮した順序でデータ投入
            from .utils.dependency import DependencyResolver
            resolver = DependencyResolver()

            # 依存関係の妥当性をチェック
            warnings = resolver.validate_dependencies(map.datamodels, schema)
            for warning in warnings:
                logger.warning(warning)

            try:
                ordered_datamodels = resolver.resolve_insertion_order(map.datamodels, schema)
            except ValueError as e:
                logger.error(f"Failed to resolve data insertion order: {e}")
                raise

            datamodels_to_process = ordered_datamodels
        else:
            # 個別指定時は従来通り
            datamodels_to_process = map.datamodels

        # データ投入処理
        for dm in datamodels_to_process:
            if dm.sync_mode == const.SYNC_MODE_MANUAL and all:
                # 手動モードで全てで指定されている場合には、スキップする
                continue
            # FIXME テーブルレイアウトが変わっている場合は、データの挿入ができない。
            if not all and target != dm.table_name:
                continue
            tbl = schema.tables[dm.table_name]
            for ds in dm.datasources:
                logger.info(f'insert {ds.filename} to {map.instance_name}.{tbl.table_name}')
                ds.load()
                table.insert(self.conn, map.instance_name, tbl, ds.data)

            if dm.sync_mode != const.SYNC_MODE_DROP_CREATE:
                # 同期モードがdrop_create以外の場合は、データのリストアを行う。
                if patch_file and target == dm.table_name:
                    # パッチファイルが指定されている場合は、パッチを実行
                    self._execute_patch(map.instance_name, tbl.table_name, patch_file)
                elif table.is_exist_backup(self.conn, map.instance_name, tbl, self.ymd):
                    # バックアップからデータを復元(同じIDは更新されるため、初期データの変更分は上書きされる)
                    logger.info(f'restore {map.instance_name}.{tbl.table_name}')
                    table.restore(self.conn, map.instance_name, tbl, self.ymd)

            engine.commit(self.conn)

    def _execute_patch(self, env: str, table_name: str, patch_file: str):
        """Execute patch file for data restoration."""
        from .patch import PatchConfig, generate_patch_sql, validate_patch_config

        try:
            # Load and validate patch configuration
            patch_config = PatchConfig.load_from_file(patch_file)

            # Validate patch configuration
            errors = validate_patch_config(patch_config)
            if errors:
                for error in errors:
                    logger.error(f"Patch validation error: {error}")
                raise ValueError(f"Invalid patch configuration: {errors[0]}")

            # Verify patch targets the correct table
            if patch_config.name != table_name:
                raise ValueError(f"Patch table name '{patch_config.name}' does not match target '{table_name}'")

            # Generate and execute SQL
            sql = generate_patch_sql(env, patch_config, self.ymd)
            logger.info(f'executing patch {patch_file} for {env}.{table_name}')
            logger.debug(f'patch SQL: {sql}')

            engine.execute(self.conn, sql)

        except Exception as e:
            logger.error(f"Failed to execute patch {patch_file}: {e}")
            raise

    # ユニットテストなどで、Operationのインスタンスを取得するためのメソッド
    @staticmethod
    def get_instance(folder: str, env: str, database: str, deploy: str):
        project = Project.load(folder)
        return Operation(project, env, database, deploy)

    def reset_all(self):
        for map in self.environ.databases:
            if self.database is not None and map.instance_name != self.database:
                continue

            self.create_database(map, 'drop')
            schema = map.build_schema(self.project.schemas, self.environ.schemas)
            self.create_table(map, schema, all, None)

    def require(self, database: str, schema_name: str, table_name: str):
        for map in self.environ.databases:
            if self.database is not None and map.instance_name != self.database:
                continue
            if map.name != database:
                continue
            schema = map.build_schema(self.project.schemas, self.environ.schemas)
            tbl = schema.tables[table_name]
            dm = map.datamodel(schema_name, table_name)
            if dm is not None:
                for ds in dm.datasources:
                    logger.info(f'insert {ds.filename} to {map.instance_name}.{tbl.table_name}')
                    ds.load()
                    table.insert(self.conn, map.instance_name, tbl, ds.data)


def apply(
        project, env: str, database: str, target: str, all: str, deploy: str,
        no_restore: bool = False, restore_only: bool = False, patch: str = None, backup_key: str = None):
    """ データベースの適用処理を行う。 CLI向け関数. """
    with Operation(project, env, database, deploy, backup_key) as op:
        for map in op.environ.databases:
            if database is not None and map.instance_name != database:
                continue
            # データベースの作成
            op.create_database(map, all)

            schema = map.build_schema(op.project.schemas, op.environ.schemas)
            op.create_table(map, schema, all, target, restore_only)
            op.insert_data(map, schema, all, target, no_restore, patch)



from graphlib import TopologicalSorter, CycleError
from logging import getLogger

logger = getLogger(__name__)


class DependencyResolver:
    """データ投入時の依存関係を解決するためのクラス"""

    def resolve_insertion_order(self, datamodels, schema):
        """
        DataModelリストを依存関係に基づいて並び替える

        Args:
            datamodels: DataModelのイテラブル
            schema: Schema オブジェクト

        Returns:
            list[DataModel]: 依存関係順に並び替えられたDataModelリスト

        Raises:
            CycleError: 循環依存が検出された場合
        """
        # DataModelをリストに変換し、キーマップを作成
        datamodels_list = list(datamodels)
        dependencies = {}
        datamodel_map = {}

        # 各DataModelの依存関係を収集
        for dm in datamodels_list:
            key = f"{dm.schema_name}@{dm.table_name}"
            datamodel_map[key] = dm

            deps = set()

            # 1. FK依存関係を抽出（スキーマから）
            if dm.table_name in schema.tables:
                table = schema.tables[dm.table_name]
                for relation in table.relations:
                    target_key = f"{relation.target.schema_name}@{relation.target.table_name}"
                    
                    # 自己参照リレーションを除外
                    if target_key == key:
                        logger.debug(f"Skipping self-reference: {key} -> {target_key}")
                        continue
                    
                    # 同じデータセット内にある依存関係のみ追加
                    if any(target_dm.schema_name == relation.target.schema_name and
                          target_dm.table_name == relation.target.table_name
                          for target_dm in datamodels_list):
                        deps.add(target_key)
                        logger.debug(f"FK dependency: {key} -> {target_key}")

            # 2. DataModelの明示的依存関係を追加
            for dep in dm.dependencies:
                # 同じデータセット内にある依存関係のみ追加
                if any(f"{target_dm.schema_name}@{target_dm.table_name}" == dep
                      for target_dm in datamodels_list):
                    deps.add(dep)
                    logger.debug(f"Explicit dependency: {key} -> {dep}")

            dependencies[key] = deps

        try:
            # TopologicalSorterで順序解決
            ts = TopologicalSorter(dependencies)
            ordered_keys = list(ts.static_order())

            # DataModelオブジェクトの順序で返す
            result = [datamodel_map[key] for key in ordered_keys if key in datamodel_map]

            logger.info(f"Resolved insertion order: {[f'{dm.schema_name}@{dm.table_name}' for dm in result]}")
            return result

        except CycleError as e:
            logger.error(f"Circular dependency detected in data insertion order: {e}")
            raise ValueError(f"Circular dependency detected: {e}")

    def validate_dependencies(self, datamodels, schema):
        """
        依存関係の妥当性を検証する

        Args:
            datamodels: DataModelのイテラブル
            schema: Schema オブジェクト

        Returns:
            list[str]: 警告メッセージのリスト
        """
        warnings = []
        datamodels_list = list(datamodels)
        available_tables = {f"{dm.schema_name}@{dm.table_name}" for dm in datamodels_list}

        for dm in datamodels_list:
            # 明示的依存関係の存在チェック
            for dep in dm.dependencies:
                if dep not in available_tables:
                    warnings.append(f"Table {dm.schema_name}@{dm.table_name} depends on {dep}, but it's not in the data insertion set")

            # FK依存関係の存在チェック
            if dm.table_name in schema.tables:
                table = schema.tables[dm.table_name]
                for relation in table.relations:
                    target_key = f"{relation.target.schema_name}@{relation.target.table_name}"
                    current_key = f"{dm.schema_name}@{dm.table_name}"
                    
                    # 自己参照リレーションを除外
                    if target_key == current_key:
                        continue
                    
                    if target_key not in available_tables:
                        warnings.append(f"Table {dm.schema_name}@{dm.table_name} has FK to {target_key}, but it's not in the data insertion set")

        return warnings
